from __future__ import unicode_literals

__title__ = 'pypuppetdb'
__version_info__ = (0, 2, 0)  # notest
__version__ = '.'.join("{0}".format(num) for num in __version_info__)  # notest
__author__ = 'Daniele Sluijters'
__license__ = 'Apache License 2.0'
__year__ = '2013, 2014, 2015'
__copyright__ = 'Copyright {0} {1}'.format(__year__, __author__)
